/**
 * \file
 * \brief ATCA Hardware abstraction layer for SAMV71 I2C over ASF drivers.
 *
 * This code is structured in two parts.  Part 1 is the connection of the ATCA HAL API to the physical I2C
 * implementation. Part 2 is the ASF I2C primitives to set up the interface.
 *
 * Prerequisite: add SERCOM I2C Master Polled support to application in Atmel Studio
 *
 * \copyright (c) 2015-2018 Microchip Technology Inc. and its subsidiaries.
 *
 * \page License
 *
 * Subject to your compliance with these terms, you may use Microchip software
 * and any derivatives exclusively with Microchip products. It is your
 * responsibility to comply with third party license terms applicable to your
 * use of third party software (including open source software) that may
 * accompany Microchip software.
 *
 * THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
 * EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
 * WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
 * PARTICULAR PURPOSE. IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT,
 * SPECIAL, PUNITIVE, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE
 * OF ANY KIND WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF
 * MICROCHIP HAS BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE
 * FORESEEABLE. TO THE FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL
 * LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED
 * THE AMOUNT OF FEES, IF ANY, THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR
 * THIS SOFTWARE.
 */

#include <string.h>
#include <stdio.h>

#include "atca_hal.h"
#include "atca_device.h"
#include "atca_execution.h"

#include "definitions.h"
#include "device.h"

/** \brief method to change the bus speed of I2C
 * \param[in] iface  interface on which to change bus speed
 * \param[in] speed  baud rate (typically 100000 or 400000)
 * \return ATCA_SUCCESS on success, otherwise an error code.
 */
ATCA_STATUS hal_i2c_change_baud(ATCAIface iface, uint32_t speed)
{
    SERCOM_I2C_TRANSFER_SETUP clk;
    clk.clkSpeed = speed;
    
    if (SERCOM6_I2C_TransferSetup(&clk, 0)) {
        return ATCA_SUCCESS;
    }
    
    return ATCA_COMM_FAIL;
}

/** \brief initialize an I2C interface using given config
 * \param[in] hal - opaque ptr to HAL data
 * \param[in] cfg - interface configuration
 * \return ATCA_SUCCESS on success, otherwise an error code.
 */
ATCA_STATUS hal_i2c_init(void *hal, ATCAIfaceCfg *cfg)
{
    // Reset the module.
    SERCOM6_REGS->I2CM.SERCOM_CTRLA = SERCOM_I2CM_CTRLA_SWRST_Msk ;

    // Wait for synchronization.
    while((SERCOM6_REGS->I2CM.SERCOM_SYNCBUSY) != 0U) {
    }

    // Enable smart mode.
    SERCOM6_REGS->I2CM.SERCOM_CTRLB = SERCOM_I2CM_CTRLB_SMEN_Msk;

    // Wait for synchronization.
    while((SERCOM6_REGS->I2CM.SERCOM_SYNCBUSY) != 0U) {
    }

    // Baud rate - Master Baud Rate.
    SERCOM6_REGS->I2CM.SERCOM_BAUD = cfg->atcai2c.baud;

    // Set Operation Mode (Master), SDA Hold time, run in stand by and i2c master enable */
    SERCOM6_REGS->I2CM.SERCOM_CTRLA = SERCOM_I2CM_CTRLA_MODE_I2C_MASTER | SERCOM_I2CM_CTRLA_SDAHOLD_75NS | SERCOM_I2CM_CTRLA_SPEED_STANDARD_AND_FAST_MODE | SERCOM_I2CM_CTRLA_SCLSM(0UL) | SERCOM_I2CM_CTRLA_ENABLE_Msk ;

    // Wait for synchronization.
    while((SERCOM6_REGS->I2CM.SERCOM_SYNCBUSY) != 0U) {
    }

    // Initial Bus State: IDLE.
    SERCOM6_REGS->I2CM.SERCOM_STATUS = (uint16_t)SERCOM_I2CM_STATUS_BUSSTATE(0x01UL);

    // Wait for synchronization.
    while((SERCOM6_REGS->I2CM.SERCOM_SYNCBUSY) != 0U) {
    }

    // Enable all Interrupts.
    SERCOM6_REGS->I2CM.SERCOM_INTENSET = (uint8_t)SERCOM_I2CM_INTENSET_Msk;

    return ATCA_SUCCESS;
}

/** \brief HAL implementation of I2C post init
 * \param[in] iface  instance
 * \return ATCA_SUCCESS
 */
ATCA_STATUS hal_i2c_post_init(ATCAIface iface)
{
    return ATCA_SUCCESS;
}

/** \brief HAL implementation of I2C send over ASF
 * \param[in] iface     instance
 * \param[in] txdata    pointer to space to bytes to send
 * \param[in] txlength  number of bytes to send
 * \return ATCA_SUCCESS on success, otherwise an error code.
 */
ATCA_STATUS hal_i2c_send(ATCAIface iface, uint8_t *txdata, int txlength)
{
    ATCAIfaceCfg *cfg = atgetifacecfg(iface);

    // txdata[0] is using _reserved byte of the ATCAPacket
    txdata[0] = 0x03;   // insert the Word Address Value, Command token
    txlength++;         // account for word address value byte.

    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }
	if (SERCOM6_I2C_Write(cfg->atcai2c.slave_address >> 1, txdata, txlength)) {
		return ATCA_SUCCESS;
	}
    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }
    
	return ATCA_COMM_FAIL;
}

/** \brief HAL implementation of I2C receive function for ASF I2C
 * \param[in]    iface     Device to interact with.
 * \param[out]   rxdata    Data received will be returned here.
 * \param[inout] rxlength  As input, the size of the rxdata buffer.
 *                         As output, the number of bytes received.
 * \return ATCA_SUCCESS on success, otherwise an error code.
 */
ATCA_STATUS hal_i2c_receive(ATCAIface iface, uint8_t *rxdata, uint16_t *rxlength)
{
    ATCAIfaceCfg *cfg = atgetifacecfg(iface);
    int retries = cfg->rx_retries;
    int status = !ATCA_SUCCESS;
    uint16_t rxdata_max_size = *rxlength;

    *rxlength = 0;
    if (rxdata_max_size < 1) {
        return ATCA_SMALL_BUFFER;
    }
    
    // Read Length byte i.e. first byte from device.
    while (retries-- > 0 && status != ATCA_SUCCESS) {
        // Busy wait until bus is free.
        while (SERCOM6_I2C_IsBusy()) {
        }
        if (SERCOM6_I2C_Read(cfg->atcai2c.slave_address >> 1, rxdata, 1) != true) {
            status = ATCA_COMM_FAIL;
        }
        else {
            status = ATCA_SUCCESS;
        }
    }

    if (status != ATCA_SUCCESS) {
        return status;
    }
    
    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }
    if (rxdata[0] < ATCA_RSP_SIZE_MIN) {
        return ATCA_INVALID_SIZE;
    }
    if (rxdata[0] > rxdata_max_size) {
        return ATCA_SMALL_BUFFER;
    }
    // Update receive length with first byte received and set to read rest of the data.
    if (SERCOM6_I2C_Read(cfg->atcai2c.slave_address >> 1, &rxdata[1], rxdata[0] - 1) != true) {
        return ATCA_COMM_FAIL;
    }
    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }

    *rxlength = rxdata[0];
    return ATCA_SUCCESS;
}

/** \brief manages reference count on given bus and releases resource if no more refences exist
 * \param[in] hal_data - opaque pointer to hal data structure - known only to the HAL implementation
 * \return ATCA_SUCCESS
 */
ATCA_STATUS hal_i2c_release(void *hal_data)
{
    return ATCA_SUCCESS;
}

/** \brief wake up CryptoAuth device using I2C bus
 * \param[in] iface  interface to logical device to wakeup
 * \return ATCA_SUCCESS on success, otherwise an error code.
 */
ATCA_STATUS hal_i2c_wake(ATCAIface iface)
{
    ATCAIfaceCfg *cfg = atgetifacecfg(iface);
    uint32_t bdrt = cfg->atcai2c.baud;
    uint8_t data[4];

    // If not already at 100kHz, change it.
    if (bdrt != 100000) {
    	hal_i2c_change_baud(iface, 100000);
    }

    // Force sleep command first in case of device reset.
    hal_i2c_sleep(iface);
    atca_delay_ms(20);

    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }
    // Send 0x00 as wake pulse.
	memset(data, 0x00, 4);
    SERCOM6_I2C_Write(0, data, 1);
    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }

    // Rounded up to the nearest ms.
	// Wait tWHI + tWLO which is configured based on device type and configuration structure.
    atca_delay_ms(((uint32_t)cfg->wake_delay + (1000 - 1)) / 1000);

    // if necessary, revert baud rate to what came in.
    if (bdrt != 100000) {
    	hal_i2c_change_baud(iface, bdrt);
    }

    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }
	SERCOM6_I2C_Read(cfg->atcai2c.slave_address >> 1, data, 4);
    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }

    // Check for wake response.
    return hal_check_wake(data, 4);
}

/** \brief idle CryptoAuth device using I2C bus
 * \param[in] iface  interface to logical device to idle
 * \return ATCA_SUCCESS on success, otherwise an error code.
 */
ATCA_STATUS hal_i2c_idle(ATCAIface iface)
{
    ATCAIfaceCfg *cfg = atgetifacecfg(iface);
    uint8_t data[4];
    
    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }
	// Idle word address value.
    data[0] = 0x02;
    if (SERCOM6_I2C_Write(cfg->atcai2c.slave_address >> 1, data, 1)) {
        return ATCA_SUCCESS;
    }
    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }
  
    return ATCA_COMM_FAIL;
}

/** \brief sleep CryptoAuth device using I2C bus
 * \param[in] iface  interface to logical device to sleep
 * \return ATCA_SUCCESS on success, otherwise an error code.
 */

ATCA_STATUS hal_i2c_sleep(ATCAIface iface)
{
    ATCAIfaceCfg *cfg = atgetifacecfg(iface);
    uint8_t data[4];

    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }
	// Sleep word address value.
    data[0] = 0x01;
    if (SERCOM6_I2C_Write(cfg->atcai2c.slave_address >> 1, data, 1)) {
        return ATCA_SUCCESS;
    }
    // Busy wait until bus is free.
    while (SERCOM6_I2C_IsBusy()) {
    }
    
    return ATCA_COMM_FAIL;
}

/** \brief discover i2c buses available for this hardware
 * this maintains a list of logical to physical bus mappings freeing the application
 * of the a-priori knowledge
 * \param[in] i2c_buses - an array of logical bus numbers
 * \param[in] max_buses - maximum number of buses the app wants to attempt to discover
 * \return ATCA_SUCCESS
 */
ATCA_STATUS hal_i2c_discover_buses(int i2c_buses[], int max_buses)
{
    return ATCA_UNIMPLEMENTED;
}

/** \brief discover any CryptoAuth devices on a given logical bus number
 * \param[in]  bus_num  Logical bus number on which to look for CryptoAuth devices
 * \param[out] cfg      Pointer to head of an array of interface config structures which get filled in by this method
 * \param[out] found    Number of devices found on this bus
 * \return ATCA_SUCCESS on success, otherwise an error code.
 */
ATCA_STATUS hal_i2c_discover_devices(int bus_num, ATCAIfaceCfg cfg[], int *found)
{
    return ATCA_UNIMPLEMENTED;
}
