/******************************************************************************
  DRV_AT24 Library Interface Implementation

  Company:
    Microchip Technology Inc.

  File Name:
    drv_at24.c

  Summary:
    AT24 EEPROM Driver Library Interface implementation

  Description:
    The AT24 Library provides a interface to access the AT24 external EEPROM.
    This file implements the AT24 Library interface.
*******************************************************************************/

//DOM-IGNORE-BEGIN
/*******************************************************************************
* Copyright (C) 2018 Microchip Technology Inc. and its subsidiaries.
*
* Subject to your compliance with these terms, you may use Microchip software
* and any derivatives exclusively with Microchip products. It is your
* responsibility to comply with third party license terms applicable to your
* use of third party software (including open source software) that may
* accompany Microchip software.
*
* THIS SOFTWARE IS SUPPLIED BY MICROCHIP "AS IS". NO WARRANTIES, WHETHER
* EXPRESS, IMPLIED OR STATUTORY, APPLY TO THIS SOFTWARE, INCLUDING ANY IMPLIED
* WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY, AND FITNESS FOR A
* PARTICULAR PURPOSE.
*
* IN NO EVENT WILL MICROCHIP BE LIABLE FOR ANY INDIRECT, SPECIAL, PUNITIVE,
* INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, COST OR EXPENSE OF ANY KIND
* WHATSOEVER RELATED TO THE SOFTWARE, HOWEVER CAUSED, EVEN IF MICROCHIP HAS
* BEEN ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE. TO THE
* FULLEST EXTENT ALLOWED BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN
* ANY WAY RELATED TO THIS SOFTWARE WILL NOT EXCEED THE AMOUNT OF FEES, IF ANY,
* THAT YOU HAVE PAID DIRECTLY TO MICROCHIP FOR THIS SOFTWARE.
*******************************************************************************/
//DOM-IGNORE-END

// *****************************************************************************
// *****************************************************************************
// Section: Include Files
// *****************************************************************************
// *****************************************************************************
#include "eeprom/drv_at24.h"

#include "definitions.h"
#include "device.h"

// *****************************************************************************
// *****************************************************************************
// Section: Global objects
// *****************************************************************************
// *****************************************************************************

/* This is the driver instance object array. */
static DRV_AT24_OBJ gDrvAT24Obj;

// *****************************************************************************
// *****************************************************************************
// Section: DRV_AT24 Driver Local Functions
// *****************************************************************************
// *****************************************************************************

/**
  * Submits a dummy I2C request to read the status of EEPROM internal write.
  * Returns true if the requested is accepted, false otherwise.
  */
static uint8_t _DRV_AT24_ReadStatus(void)
{
    gDrvAT24Obj.writeBuffer[0] = 0x00;

    /* Submit a dummy write to check internal write cycle status */
    if (SERCOM6_I2C_Write(gDrvAT24Obj.slaveAddress, gDrvAT24Obj.writeBuffer, 1) == true) {
        if (SERCOM6_I2C_ErrorGet() == SERCOM_I2C_ERROR_NONE) {
            return true;
        }
    }

    return false;
}

static bool _DRV_AT24_Write(void* txData, uint32_t txDataLength, uint32_t address)
{
    bool isRequestAccepted = false;
    uint32_t nTransferBytes = 0;
    uint32_t nBytes = 0;
    uint16_t slaveAddr;
    uint16_t i;

    if ((address + txDataLength) > gDrvAT24Obj.flashSize) {
        /* Writing past end of flash results in error */
        return isRequestAccepted;
    }

    /* Calculate the max number of bytes that can be written in the current page */
    nTransferBytes = gDrvAT24Obj.pageSize - (address % gDrvAT24Obj.pageSize);

    /* Check if the pending bytes are greater than nTransferBytes */
    nTransferBytes = txDataLength >= nTransferBytes? nTransferBytes: txDataLength;

    slaveAddr = gDrvAT24Obj.slaveAddress;
    /* Frame the EEPROM address */
    if (gDrvAT24Obj.flashSize > 131072) {    /* For 18-bit address */
    
        slaveAddr |= ((unsigned char*)&address)[2] & 0x02;
    }
    if (gDrvAT24Obj.flashSize > 65536) {     /* For 17-bit address */
        slaveAddr |= ((unsigned char*)&address)[2] & 0x01;
    }
    if (gDrvAT24Obj.flashSize > 256) {       /* For 16-bit address */
        gDrvAT24Obj.writeBuffer[nBytes++] = (address & 0x0000FF00) >> 8;
    }
    /* For 8-bit address */
    gDrvAT24Obj.writeBuffer[nBytes++] = (address & 0x000000FF);

    for (i = 0 ; i < nTransferBytes; i++) {
        gDrvAT24Obj.writeBuffer[nBytes + i] = ((uint8_t*)txData)[i];
    }

    gDrvAT24Obj.nextMemoryAddr = address + nTransferBytes;
    gDrvAT24Obj.nextBufferAddr = (uint8_t *)((uint8_t*)txData + nTransferBytes);
    gDrvAT24Obj.nPendingBytes = txDataLength - nTransferBytes;

    /* Submit a write request to I2C PLIB */
    if (SERCOM6_I2C_Write(slaveAddr, (uint8_t*)gDrvAT24Obj.writeBuffer, (nBytes + nTransferBytes)) == true) {
        
        /* Wait for EEPROM to acknowledge. */
        while (_DRV_AT24_ReadStatus() != true) {
        }
            
        /* Transfer pending bytes. */
        if (gDrvAT24Obj.nPendingBytes) {
            if (_DRV_AT24_Write(gDrvAT24Obj.nextBufferAddr, gDrvAT24Obj.nPendingBytes, gDrvAT24Obj.nextMemoryAddr) == true) {
                isRequestAccepted = true;
            }
        }
        else {
            isRequestAccepted = true;
        }
        
    }

    return isRequestAccepted;
}

// *****************************************************************************
// *****************************************************************************
// Section: DRV_AT24 Driver Global Functions
// *****************************************************************************
// *****************************************************************************

void DRV_AT24_Initialize(void)
{
    gDrvAT24Obj.slaveAddress               = 0x56;
    gDrvAT24Obj.pageSize                   = DRV_AT24_EEPROM_PAGE_SIZE;
    gDrvAT24Obj.flashSize                  = DRV_AT24_EEPROM_FLASH_SIZE;
    gDrvAT24Obj.blockStartAddress          = 0x0;
}

bool DRV_AT24_Read(void* rxData, uint32_t rxDataLength, uint32_t address)
{
    bool isRequestAccepted = false;
    uint32_t nBytes = 0;
    uint16_t slaveAddr;

    if ((rxData == NULL) || (rxDataLength == 0)) {
        return isRequestAccepted;
    }

    if ((address + rxDataLength) > gDrvAT24Obj.flashSize) {
        /* Writing past end of flash results in error */
        return isRequestAccepted;
    }

    /* Frame the EEPROM address */
    slaveAddr = gDrvAT24Obj.slaveAddress;
    if (gDrvAT24Obj.flashSize > 131072)    /* For 18-bit address */
    {
        slaveAddr |= ((unsigned char*)&address)[2] & 0x02;
    }
    if (gDrvAT24Obj.flashSize > 65536)     /* For 17-bit address */
    {
        slaveAddr |= ((unsigned char*)&address)[2] & 0x01;
    }
    if (gDrvAT24Obj.flashSize > 256)       /* For 16-bit address */
    {
        gDrvAT24Obj.writeBuffer[nBytes++] = (address & 0x0000FF00) >> 8;
    }
    /* For 8-bit address */
    gDrvAT24Obj.writeBuffer[nBytes++] = (address & 0x000000FF);

    if(SERCOM6_I2C_WriteRead(slaveAddr, (uint8_t*)gDrvAT24Obj.writeBuffer, nBytes, rxData, rxDataLength) == true) {
        isRequestAccepted = true;
    }

    return isRequestAccepted;
}

bool DRV_AT24_Write(void* txData, uint32_t txDataLength, uint32_t address)
{
    if ((txData != NULL) && (txDataLength != 0)) {
        return _DRV_AT24_Write(txData, txDataLength, address);
    }
    else {
        return false;
    }
}

bool DRV_AT24_GeometryGet(DRV_AT24_GEOMETRY *geometry)
{
    uint32_t flash_size = 0;

    flash_size = gDrvAT24Obj.flashSize;

    if ((flash_size == 0) || (gDrvAT24Obj.blockStartAddress >= flash_size)) {
        return false;
    }

    flash_size = flash_size - gDrvAT24Obj.blockStartAddress;

    /* Flash size should be at-least of a Write Block size */
    if (flash_size < gDrvAT24Obj.pageSize) {
        return false;
    }

    /* Read block size and number of blocks */
    geometry->readBlockSize = 1;
    geometry->readNumBlocks = flash_size;

    /* Write block size and number of blocks */
    geometry->writeBlockSize = gDrvAT24Obj.pageSize;
    geometry->writeNumBlocks = (flash_size / gDrvAT24Obj.pageSize);

    /* Erase block size and number of blocks */
    geometry->eraseBlockSize = 1;
    geometry->eraseNumBlocks = flash_size;

    /* Number of regions */
    geometry->readNumRegions = 1;
    geometry->writeNumRegions = 1;
    geometry->eraseNumRegions = 1;

    /* Block start address */
    geometry->blockStartAddress = gDrvAT24Obj.blockStartAddress;

    return true;
}
