/**
 * Copyright 2019-2021 EBV Elektronik. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. The name of EBV Elektronik may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EBV "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL EBV BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stddef.h>                     // Defines NULL
#include <stdbool.h>                    // Defines true
#include <stdlib.h>                     // Defines EXIT_FAILURE
#include "definitions.h"                // SYS function prototypes

/** Debug includes. */
#include "dprint/dprint.h"

/** Heracles includes. */
#include "heracles/gsm.h"
#include "heracles/gsm_socket.h"
#include "heracles/gsm_sntp.h"
#include "heracles/gsm_config.h"

/** Cryptoauthlib includes. */
#include "cryptoauthlib.h"

/** Sensors includes. */
#include "sensors.h"

/** TLS wrapper includes. */
#include "tls_socket.h"

/** MQTT client includes. */
#include "MQTTClient.h"

/** EEPROM includes. */
#include "storage.h"

/** IoTConnect includes. */
#include "iotconnect_lib.h"
#include "iotconnect_common.h"
#include "iotconnect_telemetry.h"
#include "iotconnect_discovery.h"

#define STRING_EOL    "\r\n"
#define STRING_HEADER STRING_EOL "-- EBVchips Heracles TrustFLEX IOTConnect Demo --"STRING_EOL \
"-- SAME54 Curiosity Ultra --"STRING_EOL	\
"-- Compiled: "__DATE__ " "__TIME__ " --"STRING_EOL STRING_EOL

/** Microchip ATECC608A Provisioning Type. */
#define ATECC608A_TNGTLS                    (0x6A)
#define ATECC608A_TFLXTLS                   (0x6C)
#define ATECC608A_TCSM                      (0xC0)

/** IoTConnect Demo Firmware version. */
#define IOTCONNECT_DEMO_VERSION             "1.02"

/** IoTConnect Definitions. */
#define IOTCONNECT_PROT_ENVIRONMENT         "AVNET"
#define IOTCONNECT_PROT_VERSION             "2.0"
#define IOTCONNECT_PROT_LANG                "M_C"
#define IOTCONNECT_PROT_DEVICE_ID_MAX_LEN   (19)
#define IOTCONNECT_PROT_HOST_MAX_LEN        (64)
#define IOTCONNECT_PROT_CLIENT_ID_MAX_LEN   (84)
#define IOTCONNECT_PROT_USERNAME_MAX_LEN    (256)
#define IOTCONNECT_PROT_PASSWORD_MAX_LEN    (256)
#define IOTCONNECT_PROT_PUBTOPIC_MAX_LEN    (128)
#define IOTCONNECT_PROT_SUBTOPIC_MAX_LEN    (128)
#define IOTCONNECT_DATA_CP_ID_MAX_LEN       (64)
#define IOTCONNECT_DATA_DTG_ID_MAX_LEN      (36)
#define IOTCONNECT_DATA_OTA_ID_MAX_LEN      (36)
#define IOTCONNECT_DATA_OTA_URL_MAX_LEN     (256)
#define IOTCONNECT_DATA_OTA_REASON_MAX_LEN  (256)
#define IOTCONNECT_ERR_RETRY_TIMEOUT        (15000)
#define IOTCONNECT_ERR_RETRY_CONNECT        (15)

/** IoTConnect Status Code. */
#define IOTCONNECT_STATUS_OTA_SUCCESS       (7)
#define IOTCONNECT_STATUS_OTA_PENDING       (11)
#define IOTCONNECT_STATUS_OTA_FAILED        (4)
#define IOTCONNECT_STATUS_CMD_SUCCESS       (6)
#define IOTCONNECT_STATUS_CMD_PENDING       (5)
#define IOTCONNECT_STATUS_CMD_FAILED        (4)

/** IoTConnect Domain Name to rebuild MQTT URL (reduce EEPROM storage size). */
#define IOTCONNECT_DOMAIN_NAME              ".azure-devices.net"

/** MQTT Settings. */
#define MQTT_BUFFER_SIZE                    (1400)
#define MQTT_DEVICE_REPORT_MSEC             (20000)
#define MQTT_COMMAND_TIMEOUT_MSEC           (5000)
#define MQTT_YIELD_TIMEOUT_MSEC             (500)
#define MQTT_KEEP_ALIVE_INTERVAL_SEC        (900)

/** Device OTA Status Code. */
#define DEVICE_STATUS_OTA_RESET             (0)
#define DEVICE_STATUS_OTA_DOWNLOAD          (1)

/** IoTConnect MQTTS types definition. */
typedef struct iot_prot {
	char host[IOTCONNECT_PROT_HOST_MAX_LEN];
	char clientid[IOTCONNECT_PROT_CLIENT_ID_MAX_LEN];
	char username[IOTCONNECT_PROT_USERNAME_MAX_LEN];
	char password[IOTCONNECT_PROT_PASSWORD_MAX_LEN];
	char pubtopic[IOTCONNECT_PROT_PUBTOPIC_MAX_LEN];
	char subtopic[IOTCONNECT_PROT_SUBTOPIC_MAX_LEN];
} iot_prot;
typedef struct iot_data {
	char devId[IOTCONNECT_PROT_DEVICE_ID_MAX_LEN];
	char cpId[IOTCONNECT_DATA_CP_ID_MAX_LEN];
	char templateguid[IOTCONNECT_DATA_DTG_ID_MAX_LEN + 1];
	char otaURL[IOTCONNECT_DATA_OTA_URL_MAX_LEN];
	char otaReason[IOTCONNECT_DATA_OTA_REASON_MAX_LEN];
	int32_t rc;
	int32_t ee;
	int32_t connected;
	int32_t otaActive;
	int32_t otaStatus;
	int32_t otaReport;
	iot_prot broker;
} iot_data;

/** IoTConnect MQTTS global variables. */
static iot_data g_iot_data = {0};
static uint16_t g_m_id = 0;

/** IoTConnect HTTPS global variables. */
static uint8_t g_baseURL[128];

/** MQTT client definition. */
static MQTTClient g_mqtt_client;
static uint8_t g_mqtt_rx_buffer[MQTT_BUFFER_SIZE];
static uint8_t g_mqtt_tx_buffer[MQTT_BUFFER_SIZE];

/** Debug tag prefix definition. */
static const char *TAG = "appdemo";

/** NVM data to be stored in EEPROM. */
static nvm_storage_t nvm_data;

/** SysTick counter in milliseconds. */
volatile uint32_t ms_ticks = 0;

/**
 * \brief Initialize the ATECC608A secure element. 
 */
static ATCA_STATUS initialize_atecc608a(void)
{
	int ret = ATCA_SUCCESS;
	bool lock = 0;
	uint8_t buf[ATCA_ECC_CONFIG_SIZE];

	/* ATECC608A device is defined as global variable. Set target I2C address. */
	DPRINT_I(TAG, "initialize_atecc608a: loading ATECC608A driver...");
	cfg_ateccx08a_i2c_default.atcai2c.slave_address  = ATECC608A_TFLXTLS;
	if (0 != (ret = atcab_init(&cfg_ateccx08a_i2c_default))) {
		DPRINT_E(TAG, "initialize_atecc608a: failed to start ATECC608A ERROR 0x%02x", ret);
		return ret;
	}

	if (0 != (ret = atcab_is_locked(LOCK_ZONE_DATA, &lock))) {
		DPRINT_E(TAG, "initialize_atecc608a: failed to check status ERROR 0x%02x", ret);
		return ret;
	}

	if (0 != (ret = atcab_info(buf))) {
		DPRINT_E(TAG, "initialize_atecc608a: failed to read device info ERROR 0x%02x", ret);
		return ret;
	}
	DPRINT_I(TAG, "initialize_atecc608a: read device info {%02x}{%02x}", buf[2], buf[3]);
	DPRINT_I(TAG, "initialize_atecc608a: data zone is %s", lock ? "LOCKED" : "UNLOCKED");
	return ret;
}

/**
 * \brief IoTConnect function to retrieve base URL through discovery request.
 */
static int32_t iotconnect_get_baseurl(void)
{
	char report[512];
	int rc = 0;
    uint32_t tick_count = ms_ticks;
    
    /* Open TLS socket to discovery server. */
    DPRINT_I(TAG, "Connecting to "IOTCONNECT_DISCOVERY_HOSTNAME":443...");
    if ((rc = tls_socket_open(IOTCONNECT_DISCOVERY_HOSTNAME, "443")) != MQTT_SUCCESS) {
        DPRINT_E(TAG, "Failed to connect to discovery server!\r\n");
        tls_socket_close();
        return -1;
    }        

    /* Perform HTTPS GET request. */
    rc = sprintf(report, IOTCONNECT_DISCOVERY_HEADER_TEMPLATE, g_iot_data.cpId, IOTCONNECT_PROT_ENVIRONMENT);
    if ((rc = tls_socket_write((unsigned char *)report, rc, 15000)) < 0) {
        DPRINT_E(TAG, "Failed to write request to socket!\r\n");
        tls_socket_close();
        return -1;
    }
    char *json1;
    char *json2;
    while (1) {
        memset(report, 0, sizeof(report));
        if ((rc = tls_socket_read((unsigned char *)report, sizeof(report) - 1, 15000)) < 0) {
            DPRINT_E(TAG, "Failed to read response from socket!\r\n");
            tls_socket_close();
            return -1;
        }
        json1 = strstr(report, "{\"");
        json2 = strstr(json1, "\"}");
        if (json1 && json2) {
            break;
        }
        if ((ms_ticks - tick_count) > (20000)) {
            DPRINT_E(TAG, "No valid response timeout!\r\n");
            tls_socket_close();
            return -1;
        }
    }
    tls_socket_close();

    /* Parse JSON response and save baseUrl. */
    IotclDiscoveryResponse *dr;
    dr = iotcl_discovery_parse_discovery_response((const char *)json1);
    if (dr == NULL) {
        DPRINT_E(TAG, "Parse discovery response error!\r\n");
        return -1;
    }
    memset(g_baseURL, 0, sizeof(g_baseURL));
    sprintf((char *)g_baseURL, "%s", dr->url);
    DPRINT_W(TAG, "Found discovery URL: %s", g_baseURL);
    iotcl_discovery_free_discovery_response(dr);
    return 0;
}

/**
 * \brief IoTConnect function to retrieve device specific MQTTS endpoint settings.
 */
static int32_t iotconnect_sync(void)
{
	char report[1400];
	char data[256];
    char server[128];
	int rc = 0;
    int len = 0;
    uint32_t tick_count = ms_ticks;

    /* Open TLS socket to discovery server. */
    char *path = strstr((char *)g_baseURL + 8, "/");
    memset(server, 0, sizeof(server));
    memcpy(server, (char *)g_baseURL + 8, path - (char *)(g_baseURL + 8));
    DPRINT_I(TAG, "Connecting to %s:443...", server);
    if ((rc = tls_socket_open(server, "443")) != MQTT_SUCCESS) {
        DPRINT_E(TAG, "Failed to connect to discovery server!\r\n");
        tls_socket_close();
        return -1;
    }        

    /* Perform HTTPS POST request. */
    len = sprintf(data, IOTCONNECT_DISCOVERY_PROTOCOL_POST_DATA_TEMPLATE, g_iot_data.cpId, g_iot_data.devId);
    rc = sprintf(report, IOTCONNECT_SYNC_HEADER_TEMPLATE, path, server, len, data);
    if ((rc = tls_socket_write((unsigned char *)report, rc, 15000)) < 0) {
        DPRINT_E(TAG, "Failed to write request to socket!\r\n");
        tls_socket_close();
        return -1;
    }    
    char *json1;
    char *json2;
    while (1) {
        memset(report, 0, sizeof(report));
        if ((rc = tls_socket_read((unsigned char *)report, sizeof(report) - 1, 15000)) < 0) {
            DPRINT_E(TAG, "Failed to read response from socket!\r\n");
            tls_socket_close();
            return -1;
        }
        json1 = strstr(report, "{\"");
        json2 = strstr(json1, "\"}");
        if (json1 && json2) {
            break;
        }
        if ((ms_ticks - tick_count) > (20000)) {
            DPRINT_E(TAG, "No valid response timeout!\r\n");
            tls_socket_close();
            return -1;
        }
    }
    tls_socket_close();

    /* Parse JSON response and save sync data. */
    IotclSyncResponse *sr = iotcl_discovery_parse_sync_response((const char *)json1);
    if (sr == NULL) {
        DPRINT_E(TAG, "Parse sync response error!\r\n");
        return -1;
    }
    DPRINT_W(TAG, "Found MQTT endpoint: mqtts://%s", sr->broker.host);
    strcpy(g_iot_data.broker.host, sr->broker.host);
    strcpy(g_iot_data.broker.clientid, sr->broker.client_id);
    strcpy(g_iot_data.broker.username, sr->broker.user_name);
    strcpy(g_iot_data.broker.pubtopic, sr->broker.pub_topic);
    strcpy(g_iot_data.broker.subtopic, sr->broker.sub_topic);
    strcpy(g_iot_data.templateguid, sr->dtg);
    iotcl_discovery_free_sync_response(sr);
    return 0;
}

#define IOTCONNECT_OTA_IMG_HEADER_TEMPLATE \
    "GET %s HTTP/1.1\r\n" \
    "Host: %s\r\n" \
    "User-Agent: EBV/1.00\r\n" \
    "accept: */*\r\n" \
    "\r\n"

/**
 * \brief IoTConnect function to download OTA image file.
 */
static int32_t iotconnect_download_ota_img(char *url)
{
    char data[1400];
    uint32_t data_idx = 0;
    uint32_t data_first = 0;
    char server[128];
    uint8_t page[NVMCTRL_FLASH_PAGESIZE + 1];
    uint32_t page_idx = 0;
    uint32_t page_nb = 0;
	int rc = 0;
    int resp = 0;
    int totlen = 0;
    int len = 0;
    uint32_t tick_count = ms_ticks;

    /* Open TLS socket to OTA server. */
    char *path = strstr((url + 8), "/");
    memset(server, 0, sizeof(server));
    memcpy(server, (url + 8), path - (url + 8));
    DPRINT_I(TAG, "Connecting to OTA server %s:443...", server);
    if ((rc = tls_socket_open(server, "443")) != MQTT_SUCCESS) {
		strncpy(g_iot_data.otaReason, "Failed to connect to OTA server", IOTCONNECT_DATA_OTA_REASON_MAX_LEN);
        tls_socket_close();
        return -1;
    }        

    /* Perform HTTPS GET request. */
    rc = sprintf(data, IOTCONNECT_OTA_IMG_HEADER_TEMPLATE, path, server);
    if ((rc = tls_socket_write((unsigned char *)data, rc, 15000)) < 0) {
		strncpy(g_iot_data.otaReason, "Failed to write request to OTA server", IOTCONNECT_DATA_OTA_REASON_MAX_LEN);
        tls_socket_close();
        return -1;
    }
    while (1) {
        memset(data, 0, sizeof(data));
        data_idx = 0;
        if ((rc = tls_socket_read((unsigned char *)data, sizeof(data) - 1, 15000)) < 0) {
			strncpy(g_iot_data.otaReason, "Failed to read data from OTA server", IOTCONNECT_DATA_OTA_REASON_MAX_LEN);
            tls_socket_close();
            return -1;
        }
        /* Check for 5min timeout. */
        if ((ms_ticks - tick_count) > (60000)) {
			strncpy(g_iot_data.otaReason, "Connection timeout with OTA server", IOTCONNECT_DATA_OTA_REASON_MAX_LEN);
            tls_socket_close();
            return -1;
        }
        /* Loop until data has arrived. */
        if (rc == 0) {
            continue;
        }
        /* Parse header to find content length. */
        if (resp == 0) {
            char *ptr1 = 0;
            char *ptr2 = 0;            
            if ((ptr1 = strstr(data, "Content-Length:")) != 0 && (ptr2 = strstr(ptr1 + 16, "\r\n")) != 0) {
                ptr1 += 16;
                ptr2[0] = 0;
                for (uint32_t i = 0; ptr1[i] != 0; ++i) {
                    totlen = (totlen * 10) + (ptr1[i] - '0');
                }
                DPRINT_W(TAG, "New OTA image found (%d bytes)", totlen);
                /* Binary data starts after HTML response header. */
                if ((ptr2 = strstr(ptr2 + 1, "\r\n\r\n")) != 0) {
                    ptr2 += 4;
                    data_idx = ptr2 - data;
                    rc -= data_idx;
                }
                resp = 1;
                /* Clear entire OTA memory bank. */
                DPRINT_W(TAG, "Image will be installed in memory bank %c", (NVMCTRL_StatusGet() & NVMCTRL_STATUS_AFIRST_Msk) ? 'B' : 'A');
                DPRINT_W(TAG, "Erasing OTA partition...");
                for (uint32_t block = 0; block < OTA_BANK_SIZE; block += NVMCTRL_FLASH_BLOCKSIZE) {
                    NVMCTRL_BlockErase(OTA_BANK_ADDR + block);
                    while (NVMCTRL_IsBusy()) {
                    }
                }
                memset(page, 0, NVMCTRL_FLASH_PAGESIZE + 1);
            }        
        }
        /* Increase received data size. */
        len += rc;
        DPRINT_RW("\r(%d) [%s]: Downloading... %d%%", ms_ticks, TAG, len * 100 / totlen);
        /* Prepare memory page with OTA data then write to NVM when page is full. */
        data_first = data_idx; /* Special handling for first data buffer as it might not start from offset 0! */
        while (data_idx < rc + data_first) {
            page[page_idx++] = data[data_idx++];
            if (page_idx == NVMCTRL_FLASH_PAGESIZE) {
                NVMCTRL_PageWrite((const uint32_t *)page, OTA_BANK_ADDR + NVMCTRL_FLASH_PAGESIZE * page_nb++);
                while (NVMCTRL_IsBusy()) {
                }
                memset(page, 0, NVMCTRL_FLASH_PAGESIZE + 1);
                page_idx = 0;
            }            
        }

        /* Check if transfer is complete. */
        if (len == totlen) {
            break;
        }
        tick_count = ms_ticks;
    }
    
    /* Copy remaining bytes from memory page into NVM if any. */
    if (page_idx > 0) {
        NVMCTRL_PageWrite((const uint32_t *)page, OTA_BANK_ADDR + NVMCTRL_FLASH_PAGESIZE * page_nb++);
        while (NVMCTRL_IsBusy()) {
        }
    }
    DPRINT_RW("\r");    
    tls_socket_close();
    return 0;
}

/**
 * \brief IoTConnect callback function to process incoming messages from the cloud.
 */
static void iotconnect_mqtt_callback(MessageData *msg)
{
	/* Print received MQTT message. */
	char *str = msg->message->payload;
	str[msg->message->payloadlen] = 0;
	DPRINT_W(TAG, "iotconnect_mqtt_callback: RCV |%s|\r\n", str);
    if (!iotcl_process_event(str)) {
    	DPRINT_E(TAG, "Error while processing input MQTTS data!");
    }
}

/**
 * \brief IoTConnect function to publish message to the cloud..
 */
static int32_t iotconnect_mqtt_send(const char *json, const char *topic)
{
	MQTTMessage message;
	message.qos      = QOS1;
	message.retained = 0;
	message.dup      = 0;
	message.id       = g_m_id++;
	message.payload = (void *)json;
	message.payloadlen = strlen(json);

	return MQTTPublish(&g_mqtt_client, topic, &message);
}

/**
 * \brief IoTConnect function to report twin property change to the cloud.
 */
static int32_t iotconnect_send_twin(const char *key, const char *value)
{
	int32_t ret;
	char json[MQTT_BUFFER_SIZE];

	/* Build twin property report. */
	memset(json, 0, MQTT_BUFFER_SIZE);
	sprintf(json, "{\"%s\":\"%s\"}", key, value);

	/* Send MQTT sensor report. */
	DPRINT_I(TAG, "Sending Twin update...");
	ret = iotconnect_mqtt_send(json, "$iothub/twin/PATCH/properties/reported/?$rid=1");
	return ret;
}

/**
 * \brief IoTConnect function to send OTA acknowledge to the cloud.
 */
static int32_t iotconnect_send_ota_ack(bool status, const char *msg)
{
	int32_t ret;
    uint8_t ota_id[IOTCONNECT_DATA_OTA_ID_MAX_LEN + 1];    

    memcpy(ota_id, nvm_data.otaAckId, IOTCONNECT_DATA_OTA_ID_MAX_LEN);
    ota_id[IOTCONNECT_DATA_OTA_ID_MAX_LEN] = 0;
    const char *ack_json = iotcl_create_ota_ack_response((const char *)ota_id, status, msg);

 	/* Send MQTT OTA report. */
    DPRINT_I(TAG, "Sending OTA status update...");
    ret = iotconnect_mqtt_send(ack_json, g_iot_data.broker.pubtopic);
    free((void *)ack_json);
    return ret;
}

/**
 * \brief IoTConnect function to process command received from the cloud.
 */
static void iotconnect_on_cmd(IotclEventData data)
{
    const char *command = iotcl_clone_command(data);
    DPRINT_W(TAG, "iotconnect_on_cmd: received command {%s}\r\n", command);

    /* Process led command. */
    if (!strcasecmp(command, "led on") || !strcasecmp(command, "led 10") || !strcasecmp(command, "led 11")) {
        BOARD_LED1_Clear();
    }
    if (!strcasecmp(command, "led off") || !strcasecmp(command, "led 00") || !strcasecmp(command, "led 01")) {
        BOARD_LED1_Set();
    }
    if (!strcasecmp(command, "led on") || !strcasecmp(command, "led 01") || !strcasecmp(command, "led 11")) {
        BOARD_LED2_Clear();
    }
    if (!strcasecmp(command, "led off") || !strcasecmp(command, "led 00") || !strcasecmp(command, "led 10")) {
        BOARD_LED2_Set();
    }
    free((void *)command);
    
    /* Send command acknowledge message to the cloud. */
    const char *ack_json = iotcl_create_ack_string_and_destroy_event(data, true, NULL);
    iotconnect_mqtt_send(ack_json, g_iot_data.broker.pubtopic);
    free((void *)ack_json);
}

/**
 * \brief IoTConnect function to process OTA request received from the cloud.
 */
static void iotconnect_on_ota(IotclEventData data)
{
    const char *ackid = iotcl_clone_ack_id(data);
    const char *url = iotcl_clone_download_url(data, 0);
    DPRINT_W(TAG, "iotconnect_on_ota: OTA request has arrived.");

    /* Process a valid OTA request. */
    if (ackid && url) {
        /* Save relevant OTA data in NVM to report OTA status after reset. */        
        memset(g_iot_data.otaURL, 0, IOTCONNECT_DATA_OTA_URL_MAX_LEN);
        memcpy(g_iot_data.otaURL, url, strlen(url));
        memset(nvm_data.otaAckId, 0, IOTCONNECT_DATA_OTA_ID_MAX_LEN);
        memcpy(nvm_data.otaAckId, ackid, IOTCONNECT_DATA_OTA_ID_MAX_LEN);
        nvm_data.otaActive = 1;
        if (storage_write(&nvm_data) == true) {
            /* Mark OTA has active with download status. */
            g_iot_data.otaActive = 1;
            g_iot_data.otaStatus = DEVICE_STATUS_OTA_DOWNLOAD;
            DPRINT_W(TAG, "iotconnect_on_ota: OTA process starting...");
        }
        else {
            DPRINT_E(TAG, "iotconnect_on_ota: Failed to save OTA info into NVM, OTA aborted!");
            iotconnect_send_ota_ack(false, "Failed to save OTA data in NVM");
        }
    }
    else {
    	DPRINT_E(TAG, "iotconnect_on_ota: invalid OTA data error!");
    }

    /* Free IoTConnect resources. */
    free((void *)ackid);
    free((void *)url);
}

/**
 * \brief SysTick handler used to measure accurate delay. 
 */
static void systick_count(unsigned int ctx)
{
	ms_ticks++;
}

/**
 * \brief Function to retrieve current time based on systick. 
 */
int _gettimeofday(struct timeval *tv, void *tzvp);
int _gettimeofday(struct timeval *tv, void *tzvp)
{
	tv->tv_sec = ms_ticks / 1000;
	tv->tv_usec = ( ms_ticks % 1000 ) / 1000;
	return 0;
}

/**
 * \brief Program main function. 
 */
int main (void)
{
	Network g_mqtt_network;
	MQTTPacket_connectData mqtt_options = MQTTPacket_connectData_initializer;
	uint32_t sensor_enabled = 1;
	uint32_t tick_count_report = 0;
    char curtime[25];
	int rc = 0;
	int rssi = 0;
    int retry = 0;

    /* Initialize all modules */
    SYS_Initialize ( NULL );
    BOARD_LED1_Set();
    BOARD_LED2_Set();
    
    /* Enable SysTick timer. */
    SYSTICK_TimerInitialize();
    SYSTICK_TimerCallbackSet(systick_count, 0);
    SYSTICK_TimerStart();
    
    /* Open debug port. */
	dprint_port_open();
	DPRINT_RI(STRING_HEADER);
	DPRINT_W(TAG, "EBV IoTConnect current Firmware v%s", IOTCONNECT_DEMO_VERSION);
	DPRINT_I(TAG, "EBV IoTConnect SDK v%s (%s)", IOTCONNECT_PROT_VERSION, IOTCONNECT_PROT_ENVIRONMENT);

    /* Initialize ATECC608A driver. */
	if (initialize_atecc608a() != ATCA_SUCCESS) {
		while (1);
	}

	/* Enable I2C MEMS sensors (I2C bus already initialized by ATECC608A). */
	if (Sensors_Init() != COMPONENT_OK) {
		DPRINT_W(TAG, "WARNING sensors acquisition disabled!");
		sensor_enabled = 0;
	}
    
    /* Initialize NVM (EEPROM) storage. */
    storage_init();    

    /* Initialize IoTConnect cloud agent and data structure. */
	IotclConfig iotcConfig;
	memset(&iotcConfig, 0, sizeof(iotcConfig));
	iotcConfig.device.cpid = g_iot_data.cpId;
	iotcConfig.device.duid = g_iot_data.devId;
    iotcConfig.device.env = IOTCONNECT_PROT_ENVIRONMENT;
    iotcConfig.telemetry.dtg = g_iot_data.templateguid;
    iotcConfig.event_functions.ota_cb = iotconnect_on_ota;
    iotcConfig.event_functions.cmd_cb = iotconnect_on_cmd;
    iotcl_init(&iotcConfig);
    
	/* Fetch certificates and ID from ATECC608 and display IoTConnect settings. */
	if (tls_socket_init_certificates(g_iot_data.devId, g_iot_data.cpId) < 0) {
		while (1);
	}
	sprintf(g_iot_data.broker.clientid, "%s-%s", g_iot_data.cpId, g_iot_data.devId);
	sprintf(g_iot_data.broker.pubtopic, "devices/%s/messages/events/", g_iot_data.broker.clientid);
	sprintf(g_iot_data.broker.subtopic, "devices/%s/messages/devicebound/#", g_iot_data.broker.clientid);
	DPRINT_RI("\r\n");
	DPRINT_I(TAG, ">> *** Reading IoTConnect device certificate information *** <<");
	DPRINT_I(TAG, "Company ID: {%s}", g_iot_data.cpId);
	DPRINT_I(TAG, "Device ID: {%s}\r\n", g_iot_data.devId);

	/* Initialize Heracles driver. */
	DPRINT_I(TAG, "Starting Heracles driver...");
	while (gsm_init(GSM_SIM_INTERNAL, "") != GSM_RESP_OK) {
		DPRINT_E(TAG, "Failed to start Heracles driver!");
		SYSTICK_DelayMs(1000);
	}
	DPRINT_I(TAG, "Heracles module started.\r\n");

    while (1) {
        
		/* Socket is not connected. */
		if (g_iot_data.connected == 0) {

			/* Wait for GSM network access. */
			rc = gsm_get_csq_status(&rssi, 10000);
			if (rc == GSM_RESP_TIMEOUT) {
				DPRINT_W(TAG, "Heracles timeout, reseting...");
				while (gsm_init(GSM_SIM_INTERNAL, "") != GSM_RESP_OK) {
					SYSTICK_DelayMs(1000);
				}
				continue;
			}
			else if (rc != GSM_RESP_OK) {
				DPRINT_I(TAG, "Waiting for signal...");
				SYSTICK_DelayMs(2000);
				continue;
			}
			else {
				DPRINT_I(TAG, "Current RSSI level: -%d dBm", rssi);
			}
			DPRINT_I(TAG, "Checking for GSM network...");
			if (gsm_get_reg_status(2000) != GSM_RESP_OK) {
				DPRINT_W(TAG, "No GSM network found!\r\n");
				SYSTICK_DelayMs(2000);
				continue;
			}
			else {
				DPRINT_I(TAG, "GSM network available.");
			}

			/* Wait for GPRS service attached. */
			DPRINT_I(TAG, "Checking for GPRS service...");
			if (gsm_get_gatt_status(1000) != GSM_RESP_OK) {
				DPRINT_W(TAG, "GPRS service not available!\r\n");
				SYSTICK_DelayMs(2000);
				continue;
			}
			else {
				DPRINT_I(TAG, "GPRS service attached.\r\n");
			}
            
            /* If OTA is active in download mode, fetch OTA image. */
            if (g_iot_data.otaActive && g_iot_data.otaStatus == DEVICE_STATUS_OTA_DOWNLOAD) {
                /* Download OTA image. */
                DPRINT_I(TAG, "Downloading OTA image...");
                if (iotconnect_download_ota_img(g_iot_data.otaURL) == 0) {
                    /* OTA image transfer successful, now reset system. */
                    nvm_data.otaActive = 2;
                    storage_write(&nvm_data);
                    DPRINT_W(TAG, "Device update complete!");
                    DPRINT_W(TAG, "System will now reset in 3s...\r\n");
                    SYSTICK_DelayMs(3000);
                    NVMCTRL_BankSwap();
                    while (1) {
                    }
                }
                else {
                    DPRINT_RW("\r");    
                    DPRINT_E(TAG, "Failed to download OTA image!\r\n");
                    g_iot_data.otaReport = -1;
                    nvm_data.otaActive = 0;
                    storage_write(&nvm_data);
                }
                g_iot_data.otaStatus = DEVICE_STATUS_OTA_RESET;
                g_iot_data.otaActive = 0;
                continue;
            }

            /* Retrieve MQTTS endpoint address from NVM data. */
            memset(&nvm_data, 0, sizeof(nvm_data));
            if (storage_read(&nvm_data) != true) {
                DPRINT_E(TAG, "Failed to read NVM (EEPROM) storage!");
            }
            if (storage_is_valid(&nvm_data) && retry < IOTCONNECT_ERR_RETRY_CONNECT) {
                g_iot_data.otaActive = nvm_data.otaActive;
                sprintf(g_iot_data.broker.host, "%s%s", (char *)nvm_data.host, IOTCONNECT_DOMAIN_NAME);
                memcpy(g_iot_data.templateguid, (char *)nvm_data.dtgId, IOTCONNECT_DATA_DTG_ID_MAX_LEN);
                sprintf(g_iot_data.broker.username, "%s/%s/?api-version=2018-06-30", g_iot_data.broker.host, g_iot_data.broker.clientid);
                iotcl_init(&iotcConfig);
            }
            else {
                DPRINT_W(TAG, "Device must fetch MQTTS endpoint via discovery:", g_baseURL);

                /* Use IoTConnect Discovery to retrieve base URL. */
                DPRINT_I(TAG, "Fetching IoTConnect URL to authentication server...");
                if (iotconnect_get_baseurl() < 0) {
                    DPRINT_E(TAG, "Failed to retrieve data from discovery server!");
                    SYSTICK_DelayMs(IOTCONNECT_ERR_RETRY_TIMEOUT);
                    continue;
                }

                /* Use IoTConnect Discovery to retrieve MQTT endpoint for this device. */
                DPRINT_I(TAG, "Fetching IoTConnect MQTT endpoint...");
                if (iotconnect_sync() < 0) {
                    DPRINT_E(TAG, "Failed to retrieve MQTT endpoint from server!");
                    SYSTICK_DelayMs(IOTCONNECT_ERR_RETRY_TIMEOUT);
                    continue;
                }
                
                /* Backup URL in NVM data. */
                memset((char *)nvm_data.host, 0, STORAGE_DATA_HOST_MAX_LEN);
                memcpy((char *)nvm_data.host, g_iot_data.broker.host, strlen(g_iot_data.broker.host) - strlen(IOTCONNECT_DOMAIN_NAME));
                memcpy((char *)nvm_data.dtgId, g_iot_data.templateguid, STORAGE_DATA_DTG_ID_MAX_LEN);
                storage_write(&nvm_data);
                sprintf(g_iot_data.broker.username, "%s/%s/?api-version=2018-06-30", g_iot_data.broker.host, g_iot_data.broker.clientid);
            }

			/* Prepare MQTT client to connect to IoTConnect cloud. */
			g_mqtt_network.mqttread  = &paho_mqtt_socket_read;
			g_mqtt_network.mqttwrite = &paho_mqtt_socket_write;
			g_mqtt_network.evt = 0;
			MQTTClientInit(&g_mqtt_client, &g_mqtt_network, MQTT_COMMAND_TIMEOUT_MSEC,
					g_mqtt_tx_buffer, sizeof(g_mqtt_tx_buffer),
					g_mqtt_rx_buffer, sizeof(g_mqtt_rx_buffer));

			/* Open TLS socket to MQTT server. */
			DPRINT_I(TAG, "Connecting to %s:8883...", g_iot_data.broker.host);
			if ((rc = tls_socket_open(g_iot_data.broker.host, "8883")) != MQTT_SUCCESS) {
				DPRINT_E(TAG, "Failed to connect to MQTT server!\r\n");
				SYSTICK_DelayMs(IOTCONNECT_ERR_RETRY_TIMEOUT);
                retry++;
				continue;
			}

			/* Register to MQTT server. */
			mqtt_options.keepAliveInterval = MQTT_KEEP_ALIVE_INTERVAL_SEC;
			mqtt_options.cleansession = 1;
			mqtt_options.clientID.cstring = g_iot_data.broker.clientid;
			mqtt_options.username.cstring = g_iot_data.broker.username;
			if ((rc = MQTTConnect(&g_mqtt_client, &mqtt_options)) != MQTT_SUCCESS) {
				DPRINT_E(TAG, "Failed to login to MQTT server!\r\n");
				tls_socket_close();
				SYSTICK_DelayMs(IOTCONNECT_ERR_RETRY_TIMEOUT);
                retry++;
				continue;
			}

			/* Subscribe to IOTConnect device update topic. */
			rc = MQTTSubscribe(&g_mqtt_client, g_iot_data.broker.subtopic, QOS0, &iotconnect_mqtt_callback);
			if (rc != MQTT_SUCCESS) {
				DPRINT_E(TAG, "Failed to subscribe to MQTT topic!\r\n");
				tls_socket_close();
				SYSTICK_DelayMs(IOTCONNECT_ERR_RETRY_TIMEOUT);
                retry++;
				continue;
			}

			g_iot_data.connected = 1;
            retry = 0;
			DPRINT_I(TAG, "Welcome to IoTConnect!\r\n");

			/* Report device firmware version to cloud. */
			iotconnect_send_twin("fw", IOTCONNECT_DEMO_VERSION);

			/* Initialize and synchronize clock using NTP. */
            gsm_sntp_init(0);
		}
        
		/* Socket is connected to IOTConnect. */
		if (g_iot_data.connected) {

            /* If OTA mode is enabled, disconnect from MQTT first. */
            if (g_iot_data.otaActive && g_iot_data.otaStatus == DEVICE_STATUS_OTA_DOWNLOAD) {
                tls_socket_close();
                g_iot_data.connected = 0;
                continue;
            }
            
            /* If OTA mode is enabled after reset, OTA has be done successfully. */
            if (g_iot_data.otaActive && g_iot_data.otaStatus == DEVICE_STATUS_OTA_RESET) {
                if (g_iot_data.otaActive == 2) {
                    DPRINT_W(TAG, "Now running OTA image from memory bank %c", (NVMCTRL_StatusGet() & NVMCTRL_STATUS_AFIRST_Msk) ? 'A' : 'B');
                    g_iot_data.otaReport = 1;
                    /* Blink all LEDs fast 2s to show the new OTA image is running. */
                    for (uint32_t i = 0; i < 20; ++i) {
                        BOARD_LED1_Clear();
                        BOARD_LED2_Clear();
                        SYSTICK_DelayMs(50);
                        BOARD_LED1_Set();
                        BOARD_LED2_Set();
                        SYSTICK_DelayMs(50);
                    }
                }
                else {
                    strncpy(g_iot_data.otaReason, "Error while downloading OTA image", IOTCONNECT_DATA_OTA_REASON_MAX_LEN);
                    g_iot_data.otaReport = -1;
                }
                g_iot_data.otaActive = 0;
                nvm_data.otaActive = 0;
                storage_write(&nvm_data);
            }
            
            /* Report pending OTA status if any. */
            if (g_iot_data.otaReport != 0) {
                iotconnect_send_ota_ack((g_iot_data.otaReport == 1), g_iot_data.otaReason);
                g_iot_data.otaReport = 0;
            }
            
			/* Read for incoming messages. */
			if ((rc = MQTTYield(&g_mqtt_client, MQTT_YIELD_TIMEOUT_MSEC)) != MQTT_SUCCESS) {
				DPRINT_E(TAG, "Failed to read MQTT socket!");
				DPRINT_E(TAG, "Connection lost.\r\n");
				tls_socket_close();
				g_iot_data.connected = 0;
				SYSTICK_DelayMs(IOTCONNECT_ERR_RETRY_TIMEOUT);
				continue;
			}

			/* Send sensor data to the IOTConnect cloud. */
			if ((ms_ticks - tick_count_report) > (MQTT_DEVICE_REPORT_MSEC)) {

				/* Read sensor values if ST MEMS extension is connected. */
				if (sensor_enabled) {
					Sensors_Accelero_Handler();
					Sensors_Gyro_Handler();
					Sensors_Magneto_Handler();
					Sensors_Pressure_Handler();
					Sensors_Humidity_Handler();
					Sensors_Temperature_Handler();
				}

			    /* Build IoTConnect report. */
			    IotclMessageHandle report = iotcl_telemetry_create();
                gsm_sntp_get_time(curtime);
                iotcl_telemetry_add_with_iso_time(report, curtime);
			    iotcl_telemetry_set_number(report, "acc.X", ACCELERO_Data.AXIS_X);
			    iotcl_telemetry_set_number(report, "acc.Y", ACCELERO_Data.AXIS_Y);
			    iotcl_telemetry_set_number(report, "acc.Z", ACCELERO_Data.AXIS_Z);
			    iotcl_telemetry_set_number(report, "gyro.X", GYRO_Data.AXIS_X);
			    iotcl_telemetry_set_number(report, "gyro.Y", GYRO_Data.AXIS_Y);
			    iotcl_telemetry_set_number(report, "gyro.Z", GYRO_Data.AXIS_Z);
			    iotcl_telemetry_set_number(report, "mag.X", MAGNETO_Data.AXIS_X);
			    iotcl_telemetry_set_number(report, "mag.Y", MAGNETO_Data.AXIS_Y);
			    iotcl_telemetry_set_number(report, "mag.Z", MAGNETO_Data.AXIS_Z);
			    iotcl_telemetry_set_number(report, "press", (int)PRESSURE_Data);
			    iotcl_telemetry_set_number(report, "hum", (int)HUMIDITY_Data);
			    iotcl_telemetry_set_number(report, "temp", (int)TEMPERATURE_Data);
			    const char *report_str = iotcl_create_serialized_string(report, false);

				/* Publish the sensor report. */
				DPRINT_I(TAG, "Publishing MQTT report...");
				if ((rc = iotconnect_mqtt_send(report_str, g_iot_data.broker.pubtopic)) != MQTT_SUCCESS) {
					DPRINT_E(TAG, "Failed to write MQTT socket!");
					DPRINT_E(TAG, "Connection lost.\r\n");
					tls_socket_close();
					g_iot_data.connected = 0;
                    iotcl_telemetry_destroy(report);
                    iotcl_destroy_serialized(report_str);
					SYSTICK_DelayMs(IOTCONNECT_ERR_RETRY_TIMEOUT);
					continue;
				}

			    /* Cleanup IoTConnect allocated report. */
			    iotcl_telemetry_destroy(report);
			    iotcl_destroy_serialized(report_str);

				/* Print sensor report values. */
				DPRINT_I(TAG, "---> Accelero:     %8d %8d %8d mg", ACCELERO_Data.AXIS_X, ACCELERO_Data.AXIS_Y, ACCELERO_Data.AXIS_Z);
				DPRINT_I(TAG, "---> Gyro:         %8d %8d %8d mdps", GYRO_Data.AXIS_X, GYRO_Data.AXIS_Y, GYRO_Data.AXIS_Z);
				DPRINT_I(TAG, "---> Magneto:      %8d %8d %8d mgauss", MAGNETO_Data.AXIS_X, MAGNETO_Data.AXIS_Y, MAGNETO_Data.AXIS_Z);
				DPRINT_I(TAG, "---> Pressure:     %8d hPa", (uint16_t)PRESSURE_Data);
				DPRINT_I(TAG, "---> Humidity:     %8d %%", (uint16_t)HUMIDITY_Data);
				DPRINT_I(TAG, "---> Temperature:  %8d C", (uint16_t)TEMPERATURE_Data);

				tick_count_report = ms_ticks;
				DPRINT_I(TAG, "Done.\r\n");
			}
		}
    }

    /* Execution should not come here during normal operation */

    return (EXIT_FAILURE);
}


/*******************************************************************************
 End of File
*/
