/**
 * Copyright 2019-2021 EBV Elektronik. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. The name of EBV Elektronik may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EBV "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL EBV BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include <stdlib.h>

/** ARM mbedTLS includes. */
#include "mbedtls/ssl.h"
#include "mbedtls/error.h"

/** Paho MQTT includes. */
#include "MQTTReturnCodes.h"
#include "network_interface.h"
#include "timer_interface.h"

/** TLS wrapper includes. */
#include "tls_socket.h"

/** Debug includes. */
#include "dprint/dprint.h"

/** Debug tag prefix definition. */
//static const char *TAG = "mqtt";

/**
 * \brief Read data from MQTT socket.
 *
 * \param[in] network Paho MQTT network context data.
 * \param[out] read_buffer Buffer to store data read from socket.
 * \param[in] length Length of read_buffer.
 * \param[in] timeout_ms Timeout in milliseconds to wait for data.
 * \return 0 on success, an error code otherwise.
 */
int paho_mqtt_socket_read(Network *network, unsigned char *read_buffer, size_t length, int timeout_ms)
{
    int ret;
    
    ret = tls_socket_read(read_buffer, length, timeout_ms);
    if (ret == MBEDTLS_ERR_SSL_TIMEOUT) {
        return 0;
    }
    else if (ret < 0) {
        network->evt = MQTT_DISCONNECTED_EVENT;
        return MQTT_NETWORK_DISCONNECTED_ERROR;
    }
    else {
        return ret;
    }
}

/**
 * \brief Write data to MQTT socket.
 *
 * \param[in] network Paho MQTT network context data.
 * \param[in] send_buffer Buffer to send to the socket.
 * \param[in] length Length of send_buffer.
 * \param[in] timeout_ms Timeout in milliseconds to wait for data sent.
 * \return Number of bytes sent, an error code otherwise.
 */
int paho_mqtt_socket_write(Network *network, unsigned char *send_buffer, size_t length, int timeout_ms)
{
    int ret;
    
    ret = tls_socket_write(send_buffer, length, timeout_ms);
    if (ret < 0) {
        network->evt = MQTT_DISCONNECTED_EVENT;
        return MQTT_NETWORK_DISCONNECTED_ERROR;
    }
    else {
        return ret;
    }
}
